"""Asynchronous Python client for SwitchBot."""

from collections.abc import AsyncGenerator, Generator

from aiohttp import ClientSession
from aioresponses import aioresponses
import pytest

from switchbot_api import SwitchBotAPI
from syrupy import SnapshotAssertion

from .syrupy import SwitchBotSnapshotExtension


@pytest.fixture(name="snapshot")
def snapshot_assertion(snapshot: SnapshotAssertion) -> SnapshotAssertion:
    """Return snapshot assertion fixture with the SwitchBot extension."""
    return snapshot.use_extension(SwitchBotSnapshotExtension)


@pytest.fixture
async def client() -> AsyncGenerator[SwitchBotAPI, None]:
    """Return a SwitchBot client."""
    async with (
        ClientSession() as session,
        SwitchBotAPI("token", "secret", session) as client,
    ):
        yield client


@pytest.fixture(name="responses")
def aioresponses_fixture() -> Generator[aioresponses, None, None]:
    """Return aioresponses fixture."""
    with aioresponses() as mocked_responses:
        yield mocked_responses



"""Base Commands for SwitchBot API."""

from __future__ import annotations

from enum import Enum
from typing import TypeVar


class Commands(Enum):
    """Base command class."""

    @classmethod
    def is_supported(cls, device_type: str) -> bool:
        """Is this commands supported this device type."""
        return device_type in cls.get_supported_devices()

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return []


class CommonCommands(Commands):
    """Common commands."""

    ON = "turnOn"
    OFF = "turnOff"
    TOGGLE = "toggle"
    PRESS = "press"

    # Considering the wide range of CommonCommands, they are not inherited here.
    @classmethod
    def is_supported(cls, device_type: str) -> bool:
        """Is this commands supported this device type."""
        error_msg = "CommonCommands not implement this method"
        raise NotImplementedError(error_msg)

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        error_msg = "CommonCommands not implement this method"
        raise NotImplementedError(error_msg)


class BotCommands(Commands):
    """Bot commands."""

    PRESS = "press"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Bot"]


class OthersCommands(Commands):
    """Others commands."""

    CUSTOMIZE = "customize"  # Command {user-defined button name}


class CurtainCommands(Commands):
    """Curtain & Curtain3 commands."""

    SET_POSITION = "setPosition"  # parameter(str): index0,mode0,position0  e.g. 0,ff,80
    PAUSE = "pause"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Curtain", "Curtain 3"]


class LockCommands(Commands):
    """Lock commands."""

    LOCK = "lock"
    UNLOCK = "unlock"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Smart Lock", "Smart Lock Lite", "Smart Lock Pro", "Smart Lock Ultra"]


class LockV2Commands(Commands):
    """Lock commands."""

    LOCK = "lock"
    UNLOCK = "unlock"
    DEADBOLT = "deadbolt"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Smart Lock", "Smart Lock Pro", "Smart Lock Ultra"]


class HumidifierCommands(Commands):
    """Humidifier commands."""

    # parameter: auto, set to Auto Mode
    # 101, set atomization efficiency to 34%
    # 102, set atomization efficiency to 67%
    # 103, set atomization efficiency to 100%
    SET_MODE = "setMode"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Humidifier"]


class HumidifierV2Commands(Commands):
    """Humidifier 2 commands."""

    SET_MODE = "setMode"
    SET_CHILD_LOCK = "setChildLock"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Humidifier 2"]


class AirPurifierCommands(Commands):
    """Purifier commands."""

    # Supported Device list:
    # Air Purifier VOC
    # Air Purifier Table VOC
    # Air Purifier PM2.5
    # Air Purifier Table PM2.5

    SET_MODE = "setMode"
    SET_CHILD_LOCK = "setChildLock"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return [
            "Air Purifier VOC",
            "Air Purifier Table VOC",
            "Air Purifier PM2.5",
            "Air Purifier Table PM2.5",
        ]


class AirConditionerCommands(Commands):
    """Air conditioner commands."""

    SET_ALL = "setAll"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Air Conditioner"]


class SwitchCommands(Commands):
    """Switch commands."""

    # Supported Device list:
    # Relay Switch 1
    # Relay Switch 1PM
    SET_MODE = "setMode"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Relay Switch 1", "Relay Switch 1PM"]


class Switch2PMCommands(Commands):
    """Switch commands."""

    # Supported Device list:
    # Relay Switch 2PM
    SET_MODE = "setMode"
    SET_POSITION = "setPosition"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Relay Switch 2PM"]


class RGBWLightCommands(Commands):
    """RGBWLight commands."""

    # Supported Device list:
    # Strip Light
    SET_BRIGHTNESS = "setBrightness"
    SET_COLOR = "setColor"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Strip Light"]


class RGBWWLightCommands(Commands):
    """RGBWwLight commands."""

    # Supported Device list:
    # Floor Lamp
    # Strip Light 3
    # Color Bulb
    SET_BRIGHTNESS = "setBrightness"
    SET_COLOR = "setColor"
    SET_COLOR_TEMPERATURE = "setColorTemperature"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Strip Light 3", "Floor Lamp", "Color Bulb"]


class DoorBellCommands(Commands):
    """Door Bell commands."""

    ENABLE = "enableMotionDetection"
    DISABLE = "disableMotionDetection"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Video Doorbell"]


class VacuumCommands(Commands):
    """Vacuum commands."""

    START = "start"
    STOP = "stop"
    DOCK = "dock"
    POW_LEVEL = "PowLevel"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return [
            "K10+",
            "K10+ Pro",
            "Robot Vacuum Cleaner S1",
            "Robot Vacuum Cleaner S1 Plus",
        ]


class VacuumCleanerV2Commands(Commands):
    """VacuumCleanerV2 commands."""

    # Supported Device list:
    # K20+ Pro
    # Robot Vacuum Cleaner K10+ Pro Combo
    START_CLEAN = "startClean"
    PAUSE = "pause"
    DOCK = "dock"
    SET_VOLUME = "setVolume"
    CHANGE_PARAM = "changeParam"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["K20+ Pro", "Robot Vacuum Cleaner K10+ Pro Combo"]


class VacuumCleanerV3Commands(Commands):
    """VacuumCleanerV3 commands."""

    # Supported Device list:
    # Floor Cleaning Robot S10
    # S20
    START_CLEAN = "startClean"
    PAUSE = "pause"
    DOCK = "dock"
    SET_VOLUME = "setVolume"
    CHANGE_PARAM = "changeParam"
    ADD_WATER_FOR_HUMI = "addWaterForHumi"
    SELF_CLEAN = "selfClean"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Robot Vacuum Cleaner S10", "S20"]


class CeilingLightCommands(Commands):
    """Ceiling light commands."""

    # 1-100
    SET_BRIGHTNESS = "setBrightness"
    # 2700-6500
    SET_COLOR_TEMPERATURE = "setColorTemperature"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Ceiling Light", "Ceiling Light Pro"]


class BlindTiltCommands(Commands):
    """Blind Tilt commands."""

    SET_POSITION = "setPosition"
    FULLY_OPEN = "fullyOpen"
    CLOSE_UP = "closeUp"
    CLOSE_DOWN = "closeDown"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Blind Tilt"]


class RollerShadeCommands(Commands):
    """Roller Shade commands."""

    SET_POSITION = "setPosition"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Roller Shade"]


class BatteryCirculatorFanCommands(Commands):
    """Command types for [Battery Circulator Fan] API."""

    SET_WIND_SPEED = "setWindSpeed"
    SET_WIND_MODE = "setWindMode"
    SET_NIGHT_LIGHT_MODE = "setNightLightMode"

    @classmethod
    def get_supported_devices(cls) -> list[str]:
        """Get supported devices."""
        return ["Circulator Fan", "Battery Circulator Fan"]


class TVCommands(Commands):
    """TV commands."""

    SET_CHANNEL = "SetChannel"
    VOLUME_ADD = "volumeAdd"
    VOLUME_SUB = "volumeSub"
    CHANNEL_ADD = "channelAdd"
    CHANNEL_SUB = "channelSub"


class DVDCommands(Commands):
    """DVD commands."""

    SET_MUTE = "setMute"
    FAST_FORWARD = "FastForward"
    REWIND = "Rewind"
    NEXT = "Next"
    PREVIOUS = "Previous"
    PAUSE = "Pause"
    PLAY = "Play"
    STOP = "Stop"


class SpeakerCommands(Commands):
    """Speaker commands."""

    VOLUME_ADD = "volumeAdd"
    VOLUME_SUB = "volumeSub"


class FanCommands(Commands):
    """Fan commands."""

    SWING = "swing"
    TIMER = "timer"
    LOW_SPEED = "lowSpeed"
    MIDDLE_SPEED = "middleSpeed"
    HIGH_SPEED = "highSpeed"


class LightCommands(Commands):
    """Light commands."""

    BRIGHTNESS_UP = "brightnessUp"
    BRIGHTNESS_DOWN = "brightnessDown"


T = TypeVar("T", bound=CommonCommands)



"""Asynchronous Python client for SwitchBot."""

from __future__ import annotations

from dataclasses import asdict, is_dataclass
from typing import TYPE_CHECKING, Any

from syrupy.extensions import AmberSnapshotExtension
from syrupy.extensions.amber import AmberDataSerializer

if TYPE_CHECKING:
    from syrupy.types import (
        PropertyFilter,
        PropertyMatcher,
        PropertyPath,
        SerializableData,
    )


class SwitchBotSnapshotSerializer(AmberDataSerializer):
    """SwitchBot snapshot serializer for Syrupy.

    Handles special cases for SwitchBot data structures.
    """

    @classmethod
    def _serialize(  # pylint: disable=too-many-arguments
        cls,
        data: SerializableData,
        *,
        depth: int = 0,
        exclude: PropertyFilter | None = None,
        include: PropertyFilter | None = None,
        matcher: PropertyMatcher | None = None,
        path: PropertyPath = (),
        visited: set[Any] | None = None,
    ) -> str:
        """Pre-process data before serializing.

        This allows us to handle specific cases for
        SwitchBot data structures.
        """
        serializable_data = data
        if is_dataclass(type(data)):
            serializable_data = asdict(data)

        return super()._serialize(
            serializable_data,
            depth=depth,
            exclude=exclude,
            include=include,
            matcher=matcher,
            path=path,
            visited=visited,
        )


class SwitchBotSnapshotExtension(AmberSnapshotExtension):
    """SwitchBot extension for Syrupy."""

    VERSION = "1"
    """Current version of serialization format.

    Need to be bumped when we change the SwitchBotSnapshotSerializer.
    """

    serializer_class: type[AmberDataSerializer] = SwitchBotSnapshotSerializer



"""constant for  SwitchBot API."""

from __future__ import annotations

from enum import Enum, StrEnum


class PowerState(Enum):
    """Power state."""

    ON = "on"
    OFF = "off"


class BatteryCirculatorFanMode(StrEnum):
    """Fan mode types [Battery Circulator Fan] API."""

    DIRECT = "direct"
    NATURAL = "natural"
    SLEEP = "sleep"
    BABY = "baby"


class VacuumFanSpeed(StrEnum):
    """Fan options for VacuumCommands supported devices."""

    VACUUM_FAN_SPEED_QUIET = "0"
    VACUUM_FAN_SPEED_STANDARD = "1"
    VACUUM_FAN_SPEED_STRONG = "2"
    VACUUM_FAN_SPEED_MAX = "3"


class VacuumFanSpeedV2(StrEnum):
    """Fan options for VacuumV2Commands & VacuumV3Commands supported devices."""

    VACUUM_FAN_SPEED_QUIET = "1"
    VACUUM_FAN_SPEED_STANDARD = "2"
    VACUUM_FAN_SPEED_STRONG = "3"
    VACUUM_FAN_SPEED_MAX = "4"


class VacuumCleanMode(StrEnum):
    """Clean mode for Vacuum."""

    SWEEP = "sweep"
    MOP = "mop"
    SWEEP_MOP = "sweep_mop"



"""Tests for the client."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from aiohttp import ClientError, ClientSession
from aioresponses import CallbackResult, aioresponses
import pytest

from switchbot_api import SwitchBotAPI, SwitchBotConnectionError
from tests import load_fixture
from tests.const import MOCK_URL

if TYPE_CHECKING:
    from syrupy import SnapshotAssertion


async def test_providing_session(
    responses: aioresponses,
) -> None:
    """Test putting in own session."""
    responses.get(
        f"{MOCK_URL}/devices",
        status=200,
        body=load_fixture("device_list.json"),
    )
    async with ClientSession() as session:
        client = SwitchBotAPI("abc", "def", session=session)
        await client.list_devices()
        assert client.session is not None
        assert not client.session.closed
        await client.close()
        assert not client.session.closed


async def test_using_default_session(
    responses: aioresponses,
) -> None:
    """Test creating own session."""
    responses.get(
        f"{MOCK_URL}/devices",
        status=200,
        body=load_fixture("device_list.json"),
    )
    client = SwitchBotAPI("abc", "def")
    await client.list_devices()
    assert client.session is not None
    assert not client.session.closed
    await client.close()
    assert client.session.closed


async def test_client_error(
    client: SwitchBotAPI,
    responses: aioresponses,
) -> None:
    """Test client error."""

    async def response_handler(_: str, **_kwargs: Any) -> CallbackResult:
        """Response handler for this test."""
        raise ClientError

    responses.get(
        f"{MOCK_URL}/devices",
        callback=response_handler,
    )
    with pytest.raises(SwitchBotConnectionError):
        await client.list_devices()


async def test_device_list(
    client: SwitchBotAPI,
    responses: aioresponses,
    snapshot: SnapshotAssertion,
) -> None:
    """Test listing devices."""
    responses.get(
        f"{MOCK_URL}/devices",
        status=200,
        body=load_fixture("device_list.json"),
    )
    assert await client.list_devices() == snapshot


@pytest.mark.parametrize(
    "device_fixture",
    [
        "curtain",
        "curtain3",
        "hub_2",
    ],
)
async def test_device_status(
    client: SwitchBotAPI,
    responses: aioresponses,
    snapshot: SnapshotAssertion,
    device_fixture: str,
) -> None:
    """Test fetching device status."""
    responses.get(
        f"{MOCK_URL}/devices/abc/status",
        status=200,
        body=load_fixture(f"{device_fixture}.json"),
    )
    assert await client.get_status("abc") == snapshot



"""Exceptions for the SwitchBot API."""


class SwitchBotError(Exception):
    """General SwitchBot error occurred."""


class SwitchBotConnectionError(SwitchBotError):
    """Cannot connect to the SwitchBot API."""


class SwitchBotAuthenticationError(SwitchBotError):
    """Invalid auth for the SwitchBot API."""


class SwitchBotDeviceOfflineError(SwitchBotError):
    """Device currently offline."""



"""Tools to query the SwitchBot API."""

from __future__ import annotations

import base64
from dataclasses import dataclass
import hashlib
import hmac
import logging
import socket
import time
from typing import Any, Self
import uuid

from aiohttp import ClientError, ClientResponseError, ClientSession
from aiohttp.hdrs import METH_GET, METH_POST

from switchbot_api.commands import (
    AirConditionerCommands,
    AirPurifierCommands,
    BatteryCirculatorFanCommands,
    BlindTiltCommands,
    BotCommands,
    CeilingLightCommands,
    Commands,
    CommonCommands,
    CurtainCommands,
    DoorBellCommands,
    DVDCommands,
    FanCommands,
    HumidifierCommands,
    HumidifierV2Commands,
    LightCommands,
    LockCommands,
    LockV2Commands,
    OthersCommands,
    RGBWLightCommands,
    RGBWWLightCommands,
    RollerShadeCommands,
    SpeakerCommands,
    Switch2PMCommands,
    SwitchCommands,
    T,
    TVCommands,
    VacuumCleanerV2Commands,
    VacuumCleanerV3Commands,
    VacuumCommands,
)
from switchbot_api.exceptions import (
    SwitchBotAuthenticationError,
    SwitchBotConnectionError,
    SwitchBotDeviceOfflineError,
)
from switchbot_api.models import (
    BatteryCirculatorFanMode,
    PowerState,
    VacuumCleanMode,
    VacuumFanSpeed,
    VacuumFanSpeedV2,
)

__all__ = [
    "AirConditionerCommands",
    "AirPurifierCommands",
    "BatteryCirculatorFanCommands",
    "BatteryCirculatorFanMode",
    "BlindTiltCommands",
    "BotCommands",
    "CeilingLightCommands",
    "Commands",
    "CommonCommands",
    "CurtainCommands",
    "DVDCommands",
    "DoorBellCommands",
    "FanCommands",
    "HumidifierCommands",
    "HumidifierV2Commands",
    "LightCommands",
    "LockCommands",
    "LockV2Commands",
    "OthersCommands",
    "PowerState",
    "RGBWLightCommands",
    "RGBWWLightCommands",
    "RollerShadeCommands",
    "SpeakerCommands",
    "Switch2PMCommands",
    "SwitchCommands",
    "T",
    "TVCommands",
    "VacuumCleanMode",
    "VacuumCleanerV2Commands",
    "VacuumCleanerV3Commands",
    "VacuumCommands",
    "VacuumFanSpeed",
    "VacuumFanSpeedV2",
]

_API_HOST = "https://api.switch-bot.com"

_LOGGER = logging.getLogger(__name__)
NON_OBSERVED_REMOTE_TYPES = ["Others"]


@dataclass
class Device:
    """Device."""

    device_id: str
    device_name: str
    device_type: str
    hub_device_id: str
    group: bool
    master: bool | None

    def __init__(self, **kwargs: Any) -> None:
        """Initialize."""
        self.device_id = kwargs["deviceId"]
        self.device_name = kwargs["deviceName"]
        self.device_type = kwargs.get("deviceType", "-")
        self.hub_device_id = kwargs["hubDeviceId"]
        self.group = kwargs.get("group", False)
        self.master = kwargs.get("master")


@dataclass
class Remote:
    """Remote device."""

    device_id: str
    device_name: str
    device_type: str
    hub_device_id: str

    def __init__(self, **kwargs: Any) -> None:
        """Initialize."""
        self.device_id = kwargs["deviceId"]
        self.device_name = kwargs["deviceName"]
        self.device_type = kwargs.get("remoteType", "-")
        self.hub_device_id = kwargs["hubDeviceId"]


class SwitchBotAPI:
    """SwitchBot API."""

    _close_session: bool = False

    def __init__(
        self, token: str, secret: str, session: ClientSession | None = None
    ) -> None:
        """Initialize."""
        self.token = token
        self.secret = secret
        self.session = session

    def make_headers(self, token: str, secret: str) -> dict[str, Any]:
        """Make headers."""
        nonce = uuid.uuid4()
        timestamp = int(round(time.time() * 1000))
        string_to_sign = bytes(f"{token}{timestamp}{nonce}", "utf-8")
        secret_bytes = bytes(secret, "utf-8")

        sign = base64.b64encode(
            hmac.new(
                secret_bytes, msg=string_to_sign, digestmod=hashlib.sha256
            ).digest()
        )

        return {
            "Authorization": token,
            "Content-Type": "application/json",
            "charset": "utf8",
            "t": str(timestamp),
            "sign": str(sign, "utf-8"),
            "nonce": str(nonce),
        }

    async def _request(
        self, method: str, path: str, json: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        if not self.session:
            self.session = ClientSession()
            self._close_session = True
        try:
            response = await self.session.request(
                method,
                f"{_API_HOST}/v1.1/{path}",
                headers=self.make_headers(self.token, self.secret),
                json=json,
            )
        except (
            ClientError,
            ClientResponseError,
            socket.gaierror,
        ) as exception:
            msg = "Error occurred while communicating with the SwitchBot API"
            raise SwitchBotConnectionError(msg) from exception
        if response.status == 403:
            raise SwitchBotAuthenticationError

        body = await response.json()

        if response.status >= 400:
            _LOGGER.error("Error %d: %s", response.status, body)
            raise SwitchBotConnectionError

        match body.get("statusCode"):
            case 100:
                return body["body"]  # type: ignore[no-any-return]
            case 161 | 171:
                # SwitchBot docs claim that 161 is the code for device
                # being offline, and 171 for a _hub_ being offline.
                # In testing, the Plug Mini (JP) return 171 when not
                # online too.
                raise SwitchBotDeviceOfflineError
            case _:
                _LOGGER.error("Error %d: %s", response.status, body)
                raise SwitchBotConnectionError

    async def list_devices(self) -> list[Device | Remote]:
        """List devices."""
        body = await self._request(METH_GET, "devices")
        _LOGGER.debug("Devices: %s", body)
        devices = [Device(**device) for device in body.get("deviceList")]  # type: ignore[union-attr]
        remotes = [
            Remote(**remote)
            for remote in body.get("infraredRemoteList")  # type: ignore[union-attr]
            if remote.get("remoteType") not in NON_OBSERVED_REMOTE_TYPES
        ]
        return [*devices, *remotes]

    async def get_status(self, device_id: str) -> dict[str, Any]:
        """No status for IR devices."""
        return await self._request(METH_GET, f"devices/{device_id}/status")

    async def get_webook_configuration(self) -> dict[str, Any]:
        """List webhooks."""
        json = {"action": "queryUrl"}
        return await self._request(METH_POST, "webhook/queryWebhook", json=json)

    async def setup_webhook(self, url: str) -> None:
        """Set up webhook to receive device status updates."""
        json = {"deviceList": "ALL", "action": "setupWebhook", "url": url}
        await self._request(METH_POST, "webhook/setupWebhook", json=json)

    async def delete_webhook(self, url: str) -> None:
        """Delete webhook."""
        json = {"action": "deleteWebhook", "url": url}
        await self._request(METH_POST, "webhook/deleteWebhook", json=json)

    async def send_command(
        self,
        device_id: str,
        command: T | str,
        command_type: str = "command",
        parameters: dict | str = "default",  # type: ignore[type-arg]
    ) -> None:
        """Send command to device.

        Args:
            device_id (str): The ID of the device.
            command (string | extends CommonCommands): The command to be sent.
            command_type (str, optional): The type of the command.
            Defaults to "command".
            parameters (dict | str, optional): The parameters for the command.
            Defaults to "default".

        Example JSON:
            {
                "commandType": "customize",
                "command": "ボタン", // the name of the customized button
                "parameter": "default"
            }

        """
        json = {
            "commandType": command_type,
            "command": command.value if isinstance(command, Commands) else command,
            "parameter": parameters,
        }
        await self._request(METH_POST, f"devices/{device_id}/commands", json=json)

    async def close(self) -> None:
        """Close the client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The SwitchBotAPI object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()
